/*  
    Copyright (C)  2005  Daniele Zelante
    
    This file is part of copyblock.
    
    This file is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    copyblock is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with copyblock; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef _LP64
	#define OFF_FRM "%lld"
#else
	#define OFF_FRM "%ld"
#endif


static const char * LICENSE =
"Copyright (C)  2005  Daniele Zelante \n"
" \n"    
"This program is free software; you can redistribute it and/or modify \n"
"it under the terms of the GNU General Public License as published by \n"
"the Free Software Foundation; either version 2 of the License, or \n"
"(at your option) any later version. \n"
" \n"
"This program is distributed in the hope that it will be useful, \n"
"but WITHOUT ANY WARRANTY; without even the implied warranty of \n"
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the \n"
"GNU General Public License for more details. \n"
" \n"
"You should have received a copy of the GNU General Public License \n"
"along with this program; if not, write to the Free Software \n"
"Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA \n";


static const char * ZELDAN = "\033[1;31mz\033[1;33me\033[1;32ml\033[1;36md\033[1;34ma\033[1;35mn\033[0;39m";


void banner()
{
	fprintf(
			stderr,
			"product : %s \n"
			"version : %s \n"
			"build   : %s \n"
			"URL     : http://savannah.nongnu.org/projects/%s \n"
			"author  : [%s] \n"
			"email   : <zeldan@email.it> \n"
			"license : GPLv2 \n\n%s\n"
			,
			PRODUCT,
			RELEASE,
			TIMESTAMP,
			PRODUCT,
			ZELDAN,
			LICENSE
			);
}



	
	
	


void syntax()
{
	fprintf(stderr,"syntax : %s {-r[z]|-w[s]} [-b bsize] [-m msize] [-V] filename \n",PRODUCT);
	fprintf(stderr,"-r[z]    : read from file and writes to stdout ('z' to write sparse)\n");
	fprintf(stderr,"-w[s]    : write to file reading from stdin ('s' to write sync)\n");
	fprintf(stderr,"-b bsize : set blocksize in KiB, defaults to 64 KiB \n");
	fprintf(stderr,"-m msize : set maximum size in KiB to read or write \n");
	fprintf(stderr,"-V       : display version info \n");
	fprintf(stderr,"filename : block device or regular file to work on \n");
	fprintf(stderr,"\n");
	exit(3);
}


void myfatal(const char * str)
{
	fputs(str,stderr);
	exit(2);
}
	
void myerror(const char * str)
{
	perror(str);
	exit(1);
}
		
ssize_t read_n(const int fd, char * buf, const size_t count)
{
	size_t done = 0;
	while(done<count)
	{
		ssize_t rd = read(fd,buf+done,count-done);
		if(rd==-1) return rd;
		if(rd==0) break;
		if((size_t)rd<count-done)
			fprintf(stderr,"\nwarn: slow read (%zd < %zu)\n",rd,count-done);
		done += rd;
	}
	return done;
}

ssize_t write_n(const int fd, const char * buf, const size_t count)
{
	size_t done = 0;
	while(done<count)
	{
		ssize_t wr = write(fd,buf+done,count-done);
		if(wr==-1) return wr;
		if(wr==0) break;
		if((size_t)wr < count-done) 
			fprintf(stderr,"\nwarn: slow write (%zu < %zd)\n",wr,count-done);
		done += wr;
	}
	return done;
}




off_t omin(const off_t a, const off_t b)
{
	return a < b ? a : b;
}


void printsize(const off_t size)
{
	const double sizeK = (double)size/1024.0;
	const double sizeM = sizeK/1024.0;
	const double sizeG = sizeM/1024.0;
		
	fprintf(stderr,"size = " OFF_FRM " B | %.02f KiB | %.02f MiB | %.02f GiB\n",
			size,sizeK,sizeM,sizeG);

}

void printprogress(const off_t total, const off_t part, int64_t * s0, time_t * t0)
{
	const time_t t1 = time(0);

	if(t1!=*t0)
	{
		const int progress = (int)((double)part/(double)total*100.0+0.5);
		const double speed = (double)(part-*s0)/(t1-*t0);
		const double speedK = speed/1024.0;
		const double speedM = speedK/1024.0;

		int eta = (int)((total-part) / speed);
		const int etaD = eta / (60*60*24);
		eta -= etaD * (60*60*24);
		const int etaH = eta / (60*60);
		eta -= etaH * (60*60);
		const int etaM = eta / 60;
		eta -= etaM * 60;
		const int etaS = eta;
			
		fprintf(stderr,"\r%2d%% : ETA = %dd,%02dh,%02dm,%02ds : SPEED = %.02f KiB/s | %.02f MiB/s       ",
			progress,etaD,etaH,etaM,etaS,speedK,speedM);
	
		*s0 = part;
		*t0 = t1;
	}

}

off_t fdsize(const int fd)
{
	const off_t size = lseek(fd,0,SEEK_END);
	if(size==-1) myerror("lseek (end)");
	if(-1==lseek(fd,0,SEEK_SET)) myerror("lseek (begin)");
	return size;
}

int isfilenameok(const char * filename)
{
	struct stat buf;
	if(-1==stat(filename,&buf)) myerror("stat");
	return S_ISREG(buf.st_mode) || S_ISBLK(buf.st_mode);
}

int isfileok(int fd)
{
	struct stat buf;
	if(-1==fstat(fd,&buf)) myerror("fstat");
	return S_ISREG(buf.st_mode) || S_ISBLK(buf.st_mode);
}

int isfilereg(int fd)
{
	struct stat buf;
	if(-1==fstat(fd,&buf)) myerror("fstat");
	return S_ISREG(buf.st_mode);
}




int isdirty(const char * buf, ssize_t len)
{
	ssize_t n;
	for(n=0; n<len; ++n)
		if(buf[n]) return 1;
	return 0;
}

void blockread(const char * filename, const ssize_t blocksize, const off_t maxsize, int sparse)
{
	if(!isfilenameok(filename)) myfatal("invalid file type \n");
	if(!isfilereg(1)) sparse = 0;
	
	const int fdi = open(filename,O_RDONLY);
	if(fdi==-1) myerror("open");
	
	off_t srcsize = fdsize(fdi);
	
	if(maxsize)
	{
		if(maxsize > srcsize)
			fprintf(stderr,"warn: requested size greater than real size \n");
		if(maxsize < srcsize)
		{
			fprintf(
					stderr,
					"info: truncating size from "OFF_FRM" to "OFF_FRM" \n",
					srcsize,maxsize
					);
			srcsize = maxsize;
		}
	}
		
	printsize(srcsize);
	char * buffer = (char*)malloc(blocksize);
	time_t t0 = time(0);
	off_t oldpart = 0;
	off_t part = 0;
	
	while(part<srcsize)
	{
		printprogress(srcsize,part,&oldpart,&t0);

		const ssize_t rdreq = omin(blocksize, (srcsize-part));
		const ssize_t rd = read_n(fdi, buffer, rdreq);
		if(rd==-1) myerror("\nread");
		
		if(rd<rdreq)
		{
			fprintf(stderr,"\nerror: read_n (%zd < %zd)\n",rd,rdreq);
			break;
		}

		part += rd;
		
		if(sparse && rd==blocksize && part<srcsize && !isdirty(buffer,rd))
		{
			if(-1==lseek(1,rd,SEEK_CUR)) 
				myerror("\nseek");
		}
		else
		{
			const ssize_t wr = write_n(1,buffer,rd);
			if(wr==-1) myerror("\nwrite");
			if(wr<rd)
			{
				fprintf(stderr,"\nerror: write_n (%zd < %zd)\n",wr,rd);
				break;
			}
		}
		
	}

}


void blockwrite(const char * filename, const ssize_t blocksize, const off_t maxsize, int sync)
{
	if(!isfilenameok(filename)) myfatal("invalid file type \n");
	
	int fdo = open(filename,O_RDWR|(sync ? O_SYNC : 0));
	if(fdo==-1) myerror("open");
	
	off_t devsize = fdsize(fdo);
	if(!maxsize && isfileok(0) && (devsize != fdsize(0)))
		myfatal("invalid input file size \n"); 
	
	if(maxsize && isfileok(0))
		myfatal("cannot specify maxsize when stdin is regular or block \n");

	if(maxsize)
	{
		if(maxsize > devsize)
			fprintf(stderr,"warn: requested size greater than real size \n");
		if(maxsize < devsize)
		{
			fprintf(
				stderr,
				"info: truncating size from "OFF_FRM" to "OFF_FRM" \n"
				,devsize,maxsize
				);
			devsize = maxsize;
		}
	}


	printsize(devsize);
	char * buffer = (char*)malloc(blocksize);
	time_t t0 = time(0);
	off_t oldpart = 0;
	off_t part = 0;

	while(part<devsize)
	{
		printprogress(devsize,part,&oldpart,&t0);
		
		ssize_t rdreq = omin(blocksize, (devsize-part));
		ssize_t rd = read_n(0, buffer, rdreq);
		if(rd==-1) myerror("\nread");

		if(rd<rdreq)
		{
			fprintf(stderr,"\nerror: read_n (%zd < %zd)\n",rd,rdreq);
			break;
		}

		part += rd;

		ssize_t wr = write_n(fdo, buffer, rd);
		if(wr==-1) myerror("\nwrite");
		if(wr<rd)
		{
			fprintf(stderr,"\nerror: write_n (%zd < %zd)\n",wr,rd);
			break;
		}
				
		if(rd<blocksize) break;

	}
}	


int main(int argc, char ** argv)
{
	assert(sizeof(off_t)==8);
		
	const char * optstring = "Vrwszb:m:";
	int write = 0;
	int read = 0;
	int sync = 0;
	int sparse = 0;
	ssize_t blocksize = 0;
	off_t maxsize = 0;
	
	for(;;)
	{
		const int opt = getopt(argc,argv,optstring);
		if(opt==-1) break;
		if(opt=='V') banner();
		if(opt=='r') read = 1;
		if(opt=='w') write = 1;
		if(opt=='s') sync = 1;
		if(opt=='z') sparse = 1;
		if(opt=='b')
		{
			if(blocksize!=0) syntax();
			blocksize = atoi(optarg) * 1024;
			if(blocksize==0) syntax();
		}
	
		if(opt=='m')
		{
			if(maxsize!=0) syntax();
			maxsize = ((int64_t)atoi(optarg)) * 1024;
			if(maxsize==0) syntax();
		}
	}

	if(optind+1!=argc) syntax();
	const char * filename = argv[optind];
	
	if(read==write) syntax();
	if((sparse && write) || (read && sync)) syntax();
	
	if(blocksize==0) blocksize=64*1024;


	fprintf(stderr,"%s \"%s\" ; blocksize %zd KiB",
			read ? "reading" : "writing",
			filename,
			blocksize/1024);

	if(maxsize)
		fprintf(stderr," ; maxsize "OFF_FRM" KiB",maxsize/1024);
	if(sync)
		fprintf(stderr," ; sync");
	
	fprintf(stderr,"\n");

	if(read)
		blockread(filename,blocksize,maxsize,sparse);
	if(write)
		blockwrite(filename,blocksize,maxsize,sync);

	fprintf(stderr,"OK\n");
	
	return 0;
}
	
